import numpy as np #pip install numpy
import math
import time
import sys
import os
import inspect
import types
import re

from collections import abc

from functools import wraps
import gc
import timeit


def MeasureTime(f):
    @wraps(f)
    def _wrapper(*args, **kwargs):
        gcold = gc.isenabled()
        gc.disable()
        start_time = timeit.default_timer()
        try:
            result = f(*args, **kwargs)
        finally:
            elapsed = timeit.default_timer() - start_time
            if gcold:
                gc.enable()
            print('Function "{}": {}s'.format(f.__name__, elapsed))
        return result
    return _wrapper

class MeasureBlockTime:
    def __init__(self,name="(block)", no_print = False, disable_gc = True, format_str=":.2f"):
        self.name = name
        self.no_print = no_print
        self.disable_gc = disable_gc
        self.format_str = format_str
    def __enter__(self):
        if self.disable_gc:
            self.gcold = gc.isenabled()
            gc.disable()
        self.start_time = timeit.default_timer()
        return self
    def __exit__(self,ty,val,tb):
        self.elapsed = timeit.default_timer() - self.start_time
        if self.disable_gc and self.gcold:
            gc.enable()
        if not self.no_print:
            print(('{}: {' + self.format_str + '}s').format(self.name, self.elapsed))
        return False #re-raise any exceptions

def getTime():
    return timeit.default_timer()
def getElapsedTime(start_time):
    return timeit.default_timer() - start_time

def string_to_uint8_array(bstr):
    return np.fromstring(bstr, np.uint8)
    
def string_to_float_array(bstr):
    return np.fromstring(bstr, np.float32)
    
def list_to_2d_float_array(flst, width, height):
    return np.reshape(np.asarray(flst, np.float32), (height, width))
    
def get_pfm_array(response):
    return list_to_2d_float_array(response.image_data_float, response.width, response.height)

def fill_like(val, seq):
    l = len(seq)
    if is_array_like(val) and len(val) == l:
        return val
    return [val] * len(seq)
def is_array_like(obj, string_is_array=False, tuple_is_array=True):
    result = hasattr(obj, "__len__") and hasattr(obj, '__getitem__') 
    if result and not string_is_array and isinstance(obj, (str, abc.ByteString)):
        result = False
    if result and not tuple_is_array and isinstance(obj, tuple):
        result = False
    return result
def is_scalar(x):
    return x is None or np.isscalar(x)
def is_scaler_array(x): #detects (x,y) or [x, y]
    if is_array_like(x):
        if len(x) > 0:
            return len(x) if is_scalar(x[0]) else -1
        else:
            return 0
    else:
        return -1
def is_array_of_2d_array_like(x):
    return is_array_like(x) and (len(x) == 0 or \
        (len(x) > 0 and is_2d_scaler_array_like(x[0])))
def get_public_fields(obj):
    return [attr for attr in dir(obj)
                            if not (attr.startswith("_") 
                            or inspect.isbuiltin(attr)
                            or inspect.isfunction(attr)
                            or inspect.ismethod(attr))]


    
def to_dict(obj):
    return dict([attr, getattr(obj, attr)] for attr in get_public_fields(obj))

    
def to_str(obj):
    return str(to_dict(obj))

    
def write_file(filename, bstr):
    with open(filename, 'wb') as afile:
        afile.write(bstr)

# helper method for converting getOrientation to roll/pitch/yaw
# https:#en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles
    
def has_method(o, name):
    return callable(getattr(o, name, None))
def to_eularian_angles(q):
    z = q.z_val
    y = q.y_val
    x = q.x_val
    w = q.w_val
    ysqr = y * y

    # roll (x-axis rotation)
    t0 = +2.0 * (w*x + y*z)
    t1 = +1.0 - 2.0*(x*x + ysqr)
    roll = math.atan2(t0, t1)

    # pitch (y-axis rotation)
    t2 = +2.0 * (w*y - z*x)
    if (t2 > 1.0):
        t2 = 1
    if (t2 < -1.0):
        t2 = -1.0
    pitch = math.asin(t2)

    # yaw (z-axis rotation)
    t3 = +2.0 * (w*z + x*y)
    t4 = +1.0 - 2.0 * (ysqr + z*z)
    yaw = math.atan2(t3, t4)

    return (pitch, roll, yaw)

    
def to_quaternion(pitch, roll, yaw):
    t0 = math.cos(yaw * 0.5)
    t1 = math.sin(yaw * 0.5)
    t2 = math.cos(roll * 0.5)
    t3 = math.sin(roll * 0.5)
    t4 = math.cos(pitch * 0.5)
    t5 = math.sin(pitch * 0.5)

    q = Quaternionr()
    q.w_val = t0 * t2 * t4 + t1 * t3 * t5 #w
    q.x_val = t0 * t3 * t4 - t1 * t2 * t5 #x
    q.y_val = t0 * t2 * t5 + t1 * t3 * t4 #y
    q.z_val = t1 * t2 * t4 - t0 * t3 * t5 #z
    return q

    
def wait_key(message = ''):
    ''' Wait for a key press on the console and return it. '''
    if message != '':
        print (message)

    result = None
    if os.name == 'nt':
        import msvcrt
        result = msvcrt.getch()
    else:
        import termios
        fd = sys.stdin.fileno()

        oldterm = termios.tcgetattr(fd)
        newattr = termios.tcgetattr(fd)
        newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
        termios.tcsetattr(fd, termios.TCSANOW, newattr)

        try:
            result = sys.stdin.read(1)
        except IOError:
            pass
        finally:
            termios.tcsetattr(fd, termios.TCSAFLUSH, oldterm)

    return result

    
def read_pfm(file):
    """ Read a pfm file """
    file = open(file, 'rb')

    color = None
    width = None
    height = None
    scale = None
    endian = None

    header = file.readline().rstrip()
    header = str(bytes.decode(header, encoding='utf-8'))
    if header == 'PF':
        color = True
    elif header == 'Pf':
        color = False
    else:
        raise Exception('Not a PFM file.')

    temp_str = str(bytes.decode(file.readline(), encoding='utf-8'))
    dim_match = re.match(r'^(\d+)\s(\d+)\s$', temp_str)
    if dim_match:
        width, height = map(int, dim_match.groups())
    else:
        raise Exception('Malformed PFM header.')

    scale = float(file.readline().rstrip())
    if scale < 0: # little-endian
        endian = '<'
        scale = -scale
    else:
        endian = '>' # big-endian

    data = np.fromfile(file, endian + 'f')
    shape = (height, width, 3) if color else (height, width)

    data = np.reshape(data, shape)
    # DEY: I don't know why this was there.
    #data = np.flipud(data)
    file.close()
    
    return data, scale

    
def write_pfm(file, image, scale=1):
    """ Write a pfm file """
    file = open(file, 'wb')

    color = None

    if image.dtype.name != 'float32':
        raise Exception('Image dtype must be float32.')

    image = np.flipud(image)

    if len(image.shape) == 3 and image.shape[2] == 3: # color image
        color = True
    elif len(image.shape) == 2 or len(image.shape) == 3 and image.shape[2] == 1: # greyscale
        color = False
    else:
        raise Exception('Image must have H x W x 3, H x W x 1 or H x W dimensions.')

    file.write('PF\n'.encode('utf-8')  if color else 'Pf\n'.encode('utf-8'))
    temp_str = '%d %d\n' % (image.shape[1], image.shape[0])
    file.write(temp_str.encode('utf-8'))

    endian = image.dtype.byteorder

    if endian == '<' or endian == '=' and sys.byteorder == 'little':
        scale = -scale

    temp_str = '%f\n' % scale
    file.write(temp_str.encode('utf-8'))

    image.tofile(file)

    
def write_png(filename, image):
    """ image must be numpy array H X W X channels
    """
    import zlib, struct

    buf = image.flatten().tobytes()
    width = image.shape[1]
    height = image.shape[0]

    # reverse the vertical line order and add null bytes at the start
    width_byte_4 = width * 4
    raw_data = b''.join(b'\x00' + buf[span:span + width_byte_4]
                        for span in range((height - 1) * width_byte_4, -1, - width_byte_4))

    def png_pack(png_tag, data):
        chunk_head = png_tag + data
        return (struct.pack("!I", len(data)) +
                chunk_head +
                struct.pack("!I", 0xFFFFFFFF & zlib.crc32(chunk_head)))

    png_bytes = b''.join([
        b'\x89PNG\r\n\x1a\n',
        png_pack(b'IHDR', struct.pack("!2I5B", width, height, 8, 6, 0, 0, 0)),
        png_pack(b'IDAT', zlib.compress(raw_data, 9)),
        png_pack(b'IEND', b'')])

    write_file(filename, png_bytes)

def add_windows_ctrl_c():
    def handler(a,b=None):
        sys.exit(1)
    add_windows_ctrl_c.is_handler_installed = \
        vars(add_windows_ctrl_c).setdefault('is_handler_installed',False)
    if sys.platform == "win32" and not add_windows_ctrl_c.is_handler_installed:
        if sys.stdin is not None and sys.stdin.isatty():
            #this is Console based application
            import win32api
            win32api.SetConsoleCtrlHandler(handler, True)
        #else do not install handler for non-console applications
        add_windows_ctrl_c.is_handler_installed = True




import torch.nn as nn
from tensorboardX import SummaryWriter
from torchvision import utils as tvutils

from .probe import *

class TensorboardProbe(Probe):
    def __init__(self, exp_name, run_name, epoch_config, model, 
            callbacks, metrics, log_config = Probe.LogConfig(), log_dir='d:/tlogs/'):
        super(TensorboardProbe, self).__init__(exp_name, run_name, epoch_config, model, 
            callbacks, metrics, log_config)
        self.log_writer = SummaryWriter(log_dir + exp_name + '/' + run_name)

    def on_after_batch(self, train_test, batch_state):
        super(TensorboardProbe, self).on_after_batch(train_test, batch_state)

        # dump model diagram on first batch
        if self.log_config.model_graph and self.metrics.stats.batch_index == 1:
            self.log_writer.add_graph(self.model, batch_state.input, verbose = True)

        # log false predictions
        if self.log_config.false_preds:
            pred = batch_state.output.max(1, keepdim=True)[1] # get the index of the max log-probability
            pairs = zip(input, pred.eq(label.view_as(pred)))
            incorrect =list( img for img, is_correct in pairs if not is_correct )
            img = tvutils.make_grid(incorrect)
            self.log_writer.add_image("False Pred", img, self.metrics.stats.batch_index)

    def on_after_epoch(self, test_train, loader):
        super(TensorboardProbe, self).on_after_epoch(test_train, loader)

        epoch = self.metrics.stats.epoch_index
        loss = self.metrics.stats.epoch_loss
        accuracy = self.metrics.stats.epoch_accuracy

        if self.log_config.basic:
            self.log_writer.add_scalar('loss', loss, epoch)
            self.log_writer.add_scalar('accuracy', accuracy, epoch)

        # histogram logging
        if self.log_config.param_histo_freq > 0 and self.log_config.param_histo_next_epoch == epoch:
            for name, m in self.model.named_modules():
                if isinstance(m, nn.Conv2d) or isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.Linear):
                    self.log_writer.add_histogram(name + "/" + "weight", m.weight, epoch)
                    self.log_writer.add_histogram(name + "/" + "bias", m.bias, epoch)           
            self.log_config.param_histo_next_epoch += self.log_config.param_histo_freq



import torch
import torch.nn.functional as F
import torch.nn as nn
import torch.optim as optim
import torch.optim.lr_scheduler as lr_scheduler
from .event import Event
from .grad_rat_sched import GradientRatioScheduler
from .weighted_mse_loss import WeightedMseLoss
from . import utils

class TrainTest:
    class Callbacks:
        def __init__(self):
            self.before_start = Event()
            self.before_batch = Event()
            self.before_epoch = Event()

            self.after_end = Event()
            self.after_batch = Event()
            self.after_epoch = Event()

    class BatchState:
        def __init__(self, row, output=None, loss=None, loss_all=None):
            self.row, self.output, self.loss, self.loss_all = \
                row, output, loss, loss_all
            self.input, self.target, self.in_weight, self.tar_weight = row
            self.stop_training = False

    @staticmethod
    def get_optimizer_params(model, scheduler):
        if scheduler is None or not isinstance(scheduler, str) or scheduler=='step_lr':
            return model.parameters()
        elif scheduler=='grad_rat':
            return GradientRatioScheduler.get_opt_params(model, lr)
        else:
            raise ValueError('Scheduler named {} is not supported'.format(scheduler))

    @staticmethod
    def get_scheduler(scheduler_name, optimizer, train_callbacks):
        if scheduler_name is None:
            return None
        elif scheduler_name=='step_lr':
            return lr_scheduler.MultiStepLR(optimizer, gamma=0.1, milestones=[1, 10, 100])
        elif scheduler_name=='grad_rat':
            return GradientRatioScheduler(train_callbacks, optimizer)
        else:
            raise ValueError('Scheduler named {} is not supported'.format(scheduler_name))
            
    @staticmethod
    def get_optimizer(opt_params, name, model, config, lr=None, weight_decay=None):
        lr = lr or config.train_config.lr
        weight_decay=weight_decay or config.train_config.weight_decay

        if name=='sgd':
            return optim.SGD(opt_params, lr, 
                config.train_config.momentum, weight_decay=weight_decay)
        elif name=='adam':
            return optim.Adam(opt_params, lr, weight_decay=weight_decay)
        else:
            raise ValueError('Optimizer named {} is not supported'.format(name))


    def __init__(self, model, config, loss_module, 
                 optimizer, scheduler, initializer, lr=None, weight_decay=None):

        self.train_callbacks = TrainTest.Callbacks()
        self.test_callbacks = TrainTest.Callbacks()

        self.model = model
        self.config = config
        self.train_device = config.train_config.device
        self.test_device = config.test_config.device
        
        self.param_lr = TrainTest.get_optimizer_params(self.model, scheduler)

        if isinstance(optimizer, str):
            self.optimizer = TrainTest.get_optimizer(self.param_lr, optimizer,
                model, config, lr, weight_decay)
        else:
            self.optimizer = optimizer or optim.SGD(self.param_lr,
                lr or config.train_config.lr, config.train_config.momentum)

        if isinstance(scheduler, str):
            self.scheduler = TrainTest.get_scheduler(scheduler, self.optimizer,
                self.train_callbacks)
        else:
            self.scheduler = scheduler

        if initializer is not None:
            model.apply(initializer)

        self.loss_module = loss_module or torch.nn.NLLLoss(reduction='none')

        if self.train_device == self.test_device:
            self.model.to(self.train_device)

    @staticmethod
    def get_loss(loss_module, output, target, in_weight, tar_weight):
        # For NLLLoss: output is log probability of each class, target is class ID
        # default reduction is averaging loss over each sample
        if isinstance(loss_module, WeightedMseLoss):
            loss_raw = loss_module(output, target, tar_weight) 
        else: # other losses with target weights not supported yet
            loss_raw = loss_all = loss_module(output, target) 

        # there are 3 cases
        # 1. loss_raw shape is () i.e. scaler (reduction='mean')
        # 2. loss_raw shape is (batch_size,)  (reduction='none')
        # 3. loss_raw shape is (batch_size, ...) (reduction='none' for tensor output)

        if len(loss_raw.shape) > 1: # case 3
            # sum over all except dim=0 i.e. get loss for each item in batch
            loss_all = loss_raw.sum(tuple(range(1, len(loss_raw.shape))))
            loss = loss_all.mean()
        elif len(loss_raw.shape) == 1:
            loss_all = loss_raw
            loss = loss_all.mean()
        else:
            loss_all = torch.Tensor((target.shape[0], 1))
            loss_all.fill_(loss_raw.item())
            loss = loss_raw

        return loss_all, loss

    @staticmethod
    def to_std_tuple(row):
        unpacker = lambda a0,a1=None,a2=None,a3=None:(a0,a1,a2,a3)
        input, target, in_weight, tar_weight = unpacker(*row)
        return input, target, in_weight, tar_weight

    @staticmethod
    def to_device(device, row):
        return tuple(val.to(device) if val is not None else None for val in row)

    def train_epoch(self, train_loader):
        # if train/test device is not same then we need to move model to different device each time
        if self.train_device != self.test_device:
            self.model.to(self.train_device)

        self.model.train()

        self.train_callbacks.before_epoch.notify(self, train_loader)
        
        stop_training = False
        for row in train_loader:
            row = TrainTest.to_std_tuple(row)
            batch_state = TrainTest.BatchState(row)
            self.train_callbacks.before_batch.notify(self, batch_state)
            if batch_state.stop_training:
                stop_training = True
                break

            row = input, target, in_weight, tar_weight = TrainTest.to_device(self.train_device, row)

            self.optimizer.zero_grad()

            if utils.has_method(self.model, 'fit'):
                output, loss_all, loss = self.model.fit(input, target)
            else:
                output = self.model(input)
                loss_all, loss = TrainTest.get_loss(self.loss_module, output, target, 
                    in_weight, tar_weight)

            loss.backward()
            self.optimizer.step()

            batch_state = TrainTest.BatchState(row, output, loss.item(), loss_all)
            self.train_callbacks.after_batch.notify(self, batch_state)
            if batch_state.stop_training:
                stop_training = True
                break

        self.train_callbacks.after_epoch.notify(self, train_loader)

        if self.scheduler is not None:
            self.scheduler.step()

        return stop_training

    def test_epoch(self, test_loader):
        if self.train_device != self.test_device:
            self.model.to(self.test_device)
        self.model.eval()

        self.test_callbacks.before_epoch.notify(self, test_loader)

        with torch.no_grad():
            stop_training = False
            for row in test_loader:
                row = TrainTest.to_std_tuple(row)
                batch_state = TrainTest.BatchState(row)
                self.test_callbacks.before_batch.notify(self, batch_state)
                if batch_state.stop_training:
                    stop_training = True
                    break
    
                row = input, target, in_weight, tar_weight = TrainTest.to_device(self.test_device,row)

                if utils.has_method(self.model, 'fit'):
                    output, loss_all, loss = self.model.fit(input, target)
                else:
                    output = self.model(input)
                    loss_all, loss = TrainTest.get_loss(self.loss_module, output, target, 
                        in_weight, tar_weight)

                batch_state = TrainTest.BatchState(row, output, loss.item(), loss_all)
                self.test_callbacks.after_batch.notify(self, batch_state)
                if batch_state.stop_training:
                    stop_training = True
                    break
                
        self.test_callbacks.after_epoch.notify(self, test_loader)
        return stop_training

    def fit(self, train_loader, test_loader=None, epochs=None):
        epochs = epochs or self.config.epochs

        self.train_callbacks.before_start.notify(self, epochs, train_loader)
        if test_loader is not None:
            self.test_callbacks.before_start.notify(self, epochs, test_loader)

        for epoch in range(epochs):
            stop_training = self.train_epoch(train_loader)
            if stop_training:
                break
            if test_loader is not None:
                stop_training = self.test_epoch(test_loader)
                if stop_training:
                    break

        self.train_callbacks.after_end.notify(self, epochs, train_loader)
        if test_loader is not None:
            self.test_callbacks.after_end.notify(self, epochs, test_loader)

    def find_lr(self, train_loader, start=1E-8, stop=5.5, steps=100):
        def exp_ann(i):
            lr = start * (stop/start) ** (i / (steps-1))
            return lr

        scheduler = lr_scheduler.LambdaLR(self.optimizer, exp_ann)
        self.scheduler = scheduler # allow access in callbacks
        batch_index = 0

        def on_after_batch(train_test, batch_state):
            scheduler.step()
            batch_index += 1
            if batch_index > steps:
                batch_state.stop_training = True

        self.train_callbacks.after_batch.subscribe(on_after_batch)

        self.fit(train_loader)



from .data_utils import *
from .probe import *
from .train_test import *
from .utils import *
from .dl_utils import *
from .pipeline import *
from .config import *
from .event import *
from .weighted_mse_loss import *




import torch
from torchvision import datasets, transforms

from .train_test import TrainTest
#from .probe import Probe
from .tensorwatch_probe import *
from .metrics import Metrics, ClassificationMetrics

class Pipeline:
    class TaskType:
        classification = 0
        regression = 1

    def __init__(self, config):
        self.config = config


    def find_lr(self, model, train_loader, task_type, 
                 loss_module, optimizer_name, initializer=None, 
                 start=1E-8, stop=10, steps=100):

        train_test = TrainTest(model, self.config, loss_module, 
                               optimizer, None, initializer, lr=1)

        if task_type == Pipeline.TaskType.classification:
            train_metrics = ClassificationMetrics(train_test.train_callbacks)
        elif task_type == Pipeline.TaskType.regression:
            train_metrics = Metrics(train_test.train_callbacks)
        else:
            raise NotImplementedError()

        train_probe = TensorWatchProbe('mnist_official_pipeline', 'train', self.config.train_config, 
                                 model, train_test.train_callbacks, train_metrics, self.config.log_config)

        train_test.find_lr(train_loader, start, stop, steps)


    def run(self, model, train_loader, test_loader, task_type, 
                 loss_module, optimizer, initializer=None, scheduler=None):

        train_test = TrainTest(model, self.config, loss_module,
                               optimizer, scheduler, initializer)

        if task_type == Pipeline.TaskType.classification:
            train_metrics = ClassificationMetrics(train_test.train_callbacks)
            test_metrics = ClassificationMetrics(train_test.test_callbacks)
        elif task_type == Pipeline.TaskType.regression:
            train_metrics = Metrics(train_test.train_callbacks)
            test_metrics = Metrics(train_test.test_callbacks)
        else:
            raise NotImplementedError()

        train_probe = TensorWatchProbe('mnist_official_pipeline', 'train', self.config.train_config, 
                                 model, train_test.train_callbacks, train_metrics, self.config.log_config, port=0)
        test_probe = TensorWatchProbe('mnist_official_pipeline', 'test', self.config.test_config, 
                                model, train_test.test_callbacks, test_metrics, self.config.log_config, port=1)

        train_test.fit(train_loader, test_loader)
